#!/usr/bin/env python
"""
Dimensioned constants.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from numpy import array

from aeroutils.units import unit, dimless


class PhysicalConstants():
    """Dimensionalized constants for general physics.

    :g: acceleration due to gravity
    :R: universal gas constant
    :R_air: gas constant for air
    :gamma_air: ratio of specific heats for air
    :collision_diam_air: effective collision diameter of an air molecule
    :N_A: Avogadro's constant
    :R_earth: radius of Earth
    """
    g = 9.80665 * unit('m/s^2')
    R = 8.31432 * unit('N m / (mol K)')
    R_air = 287.05287 * unit('J/(K*kg)')
    gamma_air = 1.4 * dimless
    collision_diam_air = 0.365e-9 * unit('m')
    N_A = 6.02257e23 * unit('1/mol')
    R_earth = 6.356766e3 * unit('km')


class AtmosphereConstants():
    """Dimensionalized constants for atmospheric modeling.

    :p_0: sea level ambient pressure
    :T_ice: sea level ice point temperature
    :T_0: sea level ambient temperature
    :rho_0: sea level ambient density
    :S: Sutherland's empirical constant in equation for dynamic viscosity
    :beta_s: Sutherland's empirical constant in equation for dynamic viscosity
    """

    # Sea level
    p_0 = 101.325e3 * unit('Pa')
    T_ice = 273.15 * unit('K')
    T_0 = 288.15 * unit('K')
    rho_0 = 1.225 * unit('kg/m^3')
    S = 110.4 * unit('K')
    beta_s = 1.458e-6 * unit('kg/(m*s*K^(1/2))')

    layer_names = array([
        "Troposphere",
        "Tropopause",
        "Stratosphere",
        "Stratosphere",
        "Stratopause",
        "Mesosphere",
        "Mesosphere",
        "Mesopause",
        ])

    # Layer base geopotential heights
    H_base = array([
        -5.0,
        0.0,
        11.0,
        20.0,
        32.0,
        47.0,
        51.0,
        71.0,
        80.0
        ]) * unit('km')

    # Layer temperature gradients
    T_grad = array([
        -6.5,
        -6.5,
        0.0,
        1.0,
        2.8,
        0.0,
        -2.8,
        -2.0,
        -2.0
        ]) * 1e-3 * unit('K/m')

    # Layer base temperatures
    T_base = array([
        320.65,
        288.15,
        216.65,
        216.65,
        228.65,
        270.65,
        270.65,
        214.65,
        196.65,
        ]) * unit('K')

    # Layer base pressures
    p_base = array([
        1.77687e+5,
        1.01325e+5,
        2.26320e+4,
        5.47487e+3,
        8.68014e+2,
        1.10906e+2,
        6.69384e+1,
        3.95639e+0,
        8.86272e-1,
        ]) * unit('Pa')



"""
Built-in dimensional unit checking for mathematical operations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from .units import unit, dimless, printv

__all__ = ['unit', 'dimless', 'printv']



#!/usr/bin/env python
"""Compute atmospheric properties from ICAO 1993 standard atmosphere model.

Dependencies: numpy, pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from functools import wraps
from numpy import atleast_1d, array, exp, ndarray, pi, shape, size, sqrt,\
        zeros_like

from aeroutils.constants import PhysicalConstants as Phys
from aeroutils.constants import AtmosphereConstants as Atmo
from aeroutils.units import unit
from aeroutils.units.units import check_dimensioned, check_length_dimensioned,\
        to_base_units_wrapper


def _atleast_1d(arr):
    """DEPRECATED: My version of numpy.atleast_1d that supports Python 3.7.
    This function will be deleted in future versions.

    :arr: array
    :returns: TODO

    """
    if len(shape(arr)) == 0:  # scalar, non-array
        return array([arr.magnitude]) * arr.units
    else:  # already an array
        return arr


def _len1array_to_scalar(func):
    """Decorator to output scalar if array is length 1."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        output = func(*args, **kwargs)
        if isinstance(output.magnitude, ndarray):
            if size(output) == 1:
                return output[0]
        return output
    return wrapper


def _formatarr(unitstr):
    """Decorator to format arrays from base atmospheric model.

    :arr: input array from base atmospheric model
    :unitstr: unit string
    :returns: dimensionalized, formatted array

    """
    def inner_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            arr = func(*args, **kwargs)
            arr *= unit(unitstr)
            if size(arr) == 1:
                arr = arr[0]
            return arr
        return wrapper
    return inner_decorator


class Atmosphere():

    """Compute quantities from International Civil Aviation Organization (ICAO)
    1993, which extends the US 1976 Standard Atmospheric Model to 80 km.

    Usage:

        from aeroutils import Atmosphere, unit

        # Compute atmospheric data for a scalar or array of altitudes
        h = [0.0, 12.7, 44.2, 81.0] * unit('km')
        atm = Atmosphere(h)

        # Print abbreviated output:
        print(f"\n{atm}")

        # Print extended output in Imperial units:
        print(f"\n{atm.tostring(full_output=True, imperial_units=True)}")

        # See also the linspace() function from numpy, e.g.
        # h = linspace(0, 81.0, 82) * unit('km')

        # Access individual properties and convert to desired units: "
        p, T, rho, nu, a = atm.p, atm.T, atm.rho, atm.nu, atm.a
        print(f"\nThe pressure in psi is {p.to('psi'):.5g}")

        # Compute additional properties such as thermal conductivity,
        # mean free path, and more (see class for all options)
        print(f"\nThe thermal conductivity is {atm.k:.5g}"
            f"\nThe mean free path = {atm.mean_free_path:.5g}")
    """

    class Layer():
        """Nested class to compute and store layer data."""

        def __init__(self, H_arr):
            """Initialize Layer nested class.

            :H_arr: geopotential altitude
            """
            H_arr = _atleast_1d(H_arr)

            self._name = [""]*size(H_arr)
            self._H_base = zeros_like(H_arr) * unit('m')
            self._T_base = zeros_like(H_arr) * unit('K')
            self._T_grad = zeros_like(H_arr) * unit('K/m')
            self._p_base = zeros_like(H_arr) * unit('Pa')

            for idx, H in enumerate(H_arr):
                jdx = __class__._layer_idx(H)
                self._name[idx] = Atmo.layer_names[jdx]
                self._H_base[idx] = Atmo.H_base[jdx]
                self._T_base[idx] = Atmo.T_base[jdx]
                self._T_grad[idx] = Atmo.T_grad[jdx]
                self._p_base[idx] = Atmo.p_base[jdx]

        @staticmethod
        def _layer_idx(H):
            """Find index for layer data.

            :H_arr: geopotential altitude
            :returns: index

            """
            idx = None
            for idx, H_base in enumerate(Atmo.H_base[:-1]):
                H_base_np1 = Atmo.H_base[idx+1]
                if H_base <= H < H_base_np1:
                    return idx
            else:  # no break
                raise ValueError("H out of bounds.")

        @property
        def name(self):
            """Layer name"""
            if size(self._name) == 1:
                layer_name = self._name[0]
            else:
                layer_name = self._name
            return layer_name

        @property
        @to_base_units_wrapper
        @_len1array_to_scalar
        def H_base(self):
            """Layer base geopotential height :math:`H_{base}`"""
            return self._H_base

        @property
        @to_base_units_wrapper
        @_len1array_to_scalar
        def T_base(self):
            """Layer base temperature :math:`T_{base}`"""
            return self._T_base

        @property
        @to_base_units_wrapper
        @_len1array_to_scalar
        def T_grad(self):
            """Layer base temperature gradient :math:`T_{grad}`"""
            return self._T_grad

        @property
        @to_base_units_wrapper
        @_len1array_to_scalar
        def p_base(self):
            """Layer base pressure :math:`p_{base}`"""
            return self._p_base

    def __init__(self, altitude):
        """Input altitude - object contains the corresponding atmospheric
        quantities.

        :altitude: geometric altitude input in pint dimensional units

        """

        # Compute altitude bounds
        self._H_min = Atmo.H_base[0]
        self._H_max = Atmo.H_base[-1]
        self._h_min = __class__.h_from_H(self._H_min)
        self._h_max = __class__.h_from_H(self._H_max)

        # Process altitude input
        self._h = self._process_input_altitude(altitude)
        self._H = self.H_from_h(self.h)
        self.layer = __class__.Layer(self.H)

    def __repr__(self):
        """Output string representation of class object. Default for __str__
        :returns: string output

        """
        return self.tostring(full_output=False, imperial_units=False)

    def _process_input_altitude(self, alt):
        """Check that input is of type Quantity from pint package. Check that
        input is length dimension.  Check bounds.  Format as array even if
        scalar input.

        :alt: input scalar or array of altitudes
        :returns: geometric altitude

        """
        h = _atleast_1d(alt)

        check_dimensioned(h)
        check_length_dimensioned(h)

        if len(shape(h)) > 1:
            raise TypeError("Input must be scalar or 1-D array.")

        if (h < self._h_min).any() or (self._h_max < h).any():
            raise ValueError(
                f"Input altitude is out of bounds "
                f"({self._h_min:.5g} < h < {self._h_max:.5g})"
                )

        return h

    @to_base_units_wrapper
    def knudsen_number(self, ell):
        """Compute the Knudsen number :math:`K_n`
        :returns: Knudsen number
        """
        Kn = self.mean_free_path/ell
        return Kn

    def tostring(self, full_output=True, imperial_units=False):
        """String representation of data structure.

        :full_output: set to True for full output
        :imperial_units: set to True for Imperial units
        :returns: string representation

        """
        layer_str = f"layer = {self.layer.name}"
        if imperial_units:
            h_str = f"h = {self.h.to('kft'):8.5g~P}"
            H_str = f"H = {self.H.to('kft'):8.5g~P}"
            p_str = f"p = {self.p.to('lbf/ft^2'):8.5g~P}"
            T_str = f"T = {self.T.to('degR'):8.5g~P}"
            rho_str = f"rho = {self.rho.to('slug/ft^3'):8.5g~P}"
            a_str = f"a = {self.a.to('ft/s'):8.5g~P}"
            mu_str = f"mu = {self.mu.to('lbf/ft^2 s'):8.5g~P}"
            nu_str = f"nu = {self.nu.to('ft^2/s'):8.5g~P}"
            k_str = f"k = {self.k.to('slug ft/s^3/degR'):8.5g~P}"
            g_str = f"g = {self.g.to('ft/s^2'):8.5g~P}"
            MFP_str = f"mean_free_path = {self.mean_free_path.to('ft'):8.5g~P}"
        else:  # SI units
            h_str = f"h = {self.h.to('km'):8.5g~P}"
            H_str = f"H = {self.H.to('km'):8.5g~P}"
            p_str = f"p = {self.p.to('Pa'):8.5g~P}"
            T_str = f"T = {self.T.to('degK'):8.5g~P}"
            rho_str = f"rho = {self.rho.to('kg/m^3'):8.5g~P}"
            a_str = f"a = {self.a.to('m/s'):8.5g~P}"
            mu_str = f"mu = {self.mu.to('Pa s'):8.5g~P}"
            nu_str = f"nu = {self.nu.to('m^2/s'):8.5g~P}"
            k_str = f"k = {self.k.to('W/m/K'):8.5g~P}"
            g_str = f"g = {self.g.to('m/s^2'):8.5g~P}"
            MFP_str = f"mean_free_path = {self.mean_free_path.to('m'):8.5g~P}"

        if full_output:
            repr_str = (f"{layer_str}\n"f"{h_str}\n{H_str}\n{p_str}\n{T_str}\n"
                        f"{rho_str}\n{a_str}\n{mu_str}\n{nu_str}\n{k_str}\n"
                        f"{g_str}\n{MFP_str}")
        else:
            repr_str = (f"{h_str}\n{p_str}\n{T_str}\n{rho_str}\n{a_str}\n"
                        f"{nu_str}")
        return repr_str

    @staticmethod
    def H_from_h(h):
        """Convert geometric to geopotential altitude.

        :math:`H = \\frac{R_{earth} h}{R_{earth} + h}`
        """
        R_earth = Phys.R_earth
        H = R_earth*h/(R_earth + h)
        return H

    @staticmethod
    def h_from_H(H):
        """Convert geopotential to geometric altitude.

        :math:`h = \\frac{R_{earth} H}{R_{earth} - H}`
        """
        R_earth = Phys.R_earth
        h = R_earth*H/(R_earth - H)
        return h

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def h(self):
        """Geometric height :math:`h`"""
        return self._h

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def H(self):
        """Geopotential height :math:`H`"""
        return self._H

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def p(self):
        """Air pressure :math:`p`"""
        H_base = _atleast_1d(self.layer.H_base)
        T_base = _atleast_1d(self.layer.T_base)
        T_grad = _atleast_1d(self.layer.T_grad)
        p_base = _atleast_1d(self.layer.p_base)

        H = _atleast_1d(self.H)
        T = _atleast_1d(self.T)
        g_0 = Phys.g
        R_air = Phys.R_air

        p = zeros_like(H) * unit('Pa')

        # Pressure equation changes between T_grad == 0 and T_grad != 0
        s = T_grad == 0
        p[s] = p_base[s]*exp((-g_0/(R_air*T[s]))*(H[s] - H_base[s]))

        s = T_grad != 0
        p[s] = p_base[s]*(
                1 + (T_grad[s]/T_base[s])*(H[s] - H_base[s])
                )**((1/T_grad[s])*(-g_0/R_air))

        return p

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def T(self):
        """Ambient air temperature :math:`T`"""
        T_grad = _atleast_1d(self.layer.T_grad)
        H_base = _atleast_1d(self.layer.H_base)
        T_base = _atleast_1d(self.layer.T_base)
        T = T_base + T_grad*(self.H - H_base)
        return T

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def rho(self):
        """Ambient air density :math:`\\rho`"""
        p = self.p
        T = self.T
        R_air = Phys.R_air
        rho = p/(R_air*T)
        return rho

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def a(self):
        """Ambient speed of sound :math:`a`"""
        T = self.T
        gamma_air = Phys.gamma_air
        R_air = Phys.R_air
        a = sqrt(gamma_air*R_air*T)
        return a

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def mu(self):
        """Ambient dynamic viscosity :math:`\\mu`"""
        T = self.T
        beta_s = Atmo.beta_s
        S = Atmo.S
        mu = beta_s*T**1.5/(T + S)
        return mu

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def nu(self):
        """Ambient kinematic viscosity :math:`\\nu`"""
        mu = self.mu
        rho = self.rho
        nu = mu/rho
        return nu

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def k(self):
        """Ambient thermal conductivity :math:`k`"""
        # Empirical formula requires T in Kelvin
        T_K = self.T.to('K').magnitude
        c1 = 0.002648151
        k_ = c1*T_K**1.5 / (T_K + (245.4*10**(-12/T_K)))

        # Re-assign to dimensioned variable
        k = k_ * unit('W/m/K')
        return k

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def g(self):
        """Gravitational acceleration at altitude :math:`g`"""
        h = self.h
        g_0 = Phys.g
        R_earth = Phys.R_earth
        g = g_0*(R_earth/(R_earth + h))**2
        return g

    @property
    @to_base_units_wrapper
    @_len1array_to_scalar
    def mean_free_path(self):
        """Mean free path"""
        p = self.p
        T = self.T
        N_A = Phys.N_A
        R = Phys.R
        sigma = Phys.collision_diam_air

        n = N_A*p/(R*T)  # number density
        mean_free_path = 1/(sqrt(2)*pi*sigma**2*n)
        return mean_free_path



"""
Module to easily convert between Mach number, true airspeed (TAS), calibrated
airspeed (CAS), and equivalent airspeed (EAS) for given altitude(s).
Additional flight condition data and atmospheric data is computed.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from .flightcondition import FlightCondition



#!/usr/bin/env python
"""Provide dimensional units for mathematical operations.
Using the pint package.

Dependencies: pint

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from functools import wraps
from inspect import currentframe

from pint import UnitRegistry

unit = UnitRegistry(system='mks')
unit.default_format = '~P'
dimless = unit('dimensionless')


def check_dimensioned(inp):
    """Check that input is dimensional (type 'Quantity' from pint package) and
    *not* from a different unit registry.  """

    if isinstance(inp, unit.Quantity):
        dummy = 1 * unit('m')
        # Registry equality check
        if inp._REGISTRY is not dummy._REGISTRY:
            msg = ("Cannot operate with {} and {} of different registries."
                   "For units, try:\n\tfrom aeroutils import unit ")
            raise ValueError(
                msg.format(inp.__class__.__name__, dummy.__class__.__name__)
            )
    else:
        raise TypeError("Input value is not correctly typed! Use"
                        " dimensional type 'Quantity' from pint package.")


def check_length_dimensioned(inp):
    """Check that input is length dimension type Quantity from pint package."""
    length_dimensionality = (1*unit('ft')).dimensionality
    if not (inp.dimensionality == length_dimensionality):
        raise TypeError("Input value is not correctly typed! Use length"
                        " dimensional unit.")


def check_area_dimensioned(inp):
    """Check that input is area type Quantity from pint package."""
    area_dimensionality = (1*unit('ft^2')).dimensionality
    if not (inp.dimensionality == area_dimensionality):
        raise TypeError("Input value is not correctly typed! Use area"
                        " dimensional unit.")


def name_of_var(var):
    """Find name of variables using local items.

    :var: dimensioned variable
    :returns: user-coded name of variable

    """
    try:
        local_vars = currentframe().f_back.f_back.f_locals.items()
        match = [name for name, val in local_vars if val is var]
    except AttributeError:
        local_vars = currentframe().f_back.f_locals.items()
        match = [name for name, val in local_vars if val is var]
    name = match[0] if match else "unknown"
    return name


def printv(var, to=None, var_name="", *args, **kwargs):
    """Print name and value of a Pint unit-specified variable.
    For example,

        distance = 99.9 * unit('m')
        printv(distance)
        # prints "distance = 99.9 m"

    *Note*: as of Python 3.8, simply use the f-string syntax, e.g.
        x=7
        print(f"{x=}")

    :var: variable to be printed
    :to: (str), convert to another unit
    :var_name: overwrite variable name
    :*args: additional arguments
    :**kwargs: additional keyword arguments
    :returns: None

    """
    formatted_output = (var.to(to) if to is not None else var.to_base_units())
    var_name = var_name if var_name else name_of_var(var)
    print(f"{var_name} = {formatted_output:.5g~P}", *args, **kwargs)


def to_base_units_wrapper(func):
    """Function decorator to convert output variable units to base units. """
    @wraps(func)
    def wrapper(*args, **kwargs):
        output = func(*args, **kwargs)
        output.ito_base_units()
        return output
    return wrapper



#!/usr/bin/env python
"""Easily convert between Mach number, true airspeed (TAS), calibrated airspeed
(CAS), and equivalent airspeed (EAS) for given altitude(s).  Additional flight
condition data and atmospheric data is computed.

Dependencies: numpy, pint, scipy

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from numpy import atleast_1d, ones, sqrt, shape

from aeroutils.atmosphere import Atmosphere
from aeroutils.atmosphere.atmosphere import _atleast_1d  # DEPRECATED
from aeroutils.constants import PhysicalConstants as Phys
from aeroutils.units import unit, dimless
from aeroutils.units.units import check_dimensioned, check_area_dimensioned,\
        check_length_dimensioned, to_base_units_wrapper


class FlightCondition:

    """Easily convert between Mach number, true airspeed (TAS), calibrated
    airspeed (CAS), and equivalent airspeed (EAS) for given altitude(s).
    Additional flight condition data and atmospheric data is computed.

    All inputs must be dimensional unit quantities.

    Usage:

        from aeroutils import FlightCondition, unit, dimless

        # Compute flight conditions for a scalar or array of altitudes
        altitudes = [0, 10e3, 33.5e3] * unit('ft')
        fc = FlightCondition(altitudes, EAS=300*unit('knots'))

        # Print flight condition data:
        print(f"{fc}")

        # Print extended output in Imperial units:
        print(f"\n{fc.tostring(full_output=True, imperial_units=True)}")

        # Access flight speed formats individually
        M_inf, U_inf, U_CAS, U_EAS = fc.mach, fc.TAS, fc.CAS, fc.EAS

        # Access atmospheric data alone (see Atmosphere class for options)
        atm = fc.atm  # access Atmosphere object 'atm'
        p, T, rho, nu, a = atm.p, atm.T, atm.rho, atm.nu, atm.a

        # Input true/calibrated/equivalent airspeed or Mach number
        fc_TAS = FlightCondition(altitudes, TAS=300*unit('knots'))
        fc_CAS = FlightCondition(altitudes, CAS=300*unit('knots'))
        fc_EAS = FlightCondition(altitudes, EAS=300*unit('knots'))
        fc_mach = FlightCondition(altitudes, mach=0.4535*dimless)

        # Specify desired units on input and output
        altitudes_in_km = [0, 3.048, 10.2108] * unit('km')
        fc_other_units = FlightCondition(altitudes, EAS=154.33*unit('m/s'))
        U_TAS = fc_other_units.TAS
        print(f"\nThe true airspeed in m/s is {U_TAS.to('m/s'):.5g}")
        print(f"The true airspeed in km/s is {U_TAS.to('km/s'):.5g}")

        # Compute additional derived quantities (see class for all options)
        print(f"\nThe dynamic pressure in psi is {fc.q_inf.to('psi'):.5g}")
        ell = 60 * unit('in')  # arbitrary length scale of interest
        print(f"The Reynolds number is {fc.reynolds_number(ell):.5g}")
        print(f"The Reynolds number per-unit-length [1/in] is "
            f"{fc.reynolds_number_per_unit_length('in'):.5g}")
    """

    def __init__(self, altitude, mach=None, TAS=None, CAS=None, EAS=None,):
        """Constructor based on altitude and input speed in terms of Mach
        number, TAS, CAS, or EAS.  At least one input speed is required.  All
        inputs must be dimensional unit quantities.

        :altitude: geometric altitude
        :mach: Mach number
        :TAS: true airspeed
        :CAS: calibrated airspeed
        :EAS: equivalent airspeed

        """

        # Automatically process altitude through Atmosphere class
        self.atm = Atmosphere(altitude)
        self.altitude = self.atm.h

        h0 = 0 * unit('kft')
        self._atm0 = Atmosphere(h0)

        p_inf = self.atm.p
        p_inf_h0 = self._atm0.p
        self.delta = p_inf/p_inf_h0

        if mach is not None:
            self.mach = self.checkandsize(mach)
            self.TAS = self.TAS_from_M_inf(self.mach)
            self.EAS = self.EAS_from_TAS(self.TAS, self.mach)
            self.q_c = self.q_c_from_M_inf(self.mach)
            self.CAS = self.CAS_from_q_c(self.q_c)
        elif TAS is not None:
            self.TAS = self.checkandsize(TAS)
            self.mach = self.M_inf_from_TAS(TAS)
            self.EAS = self.EAS_from_TAS(self.TAS, self.mach)
            self.q_c = self.q_c_from_M_inf(self.mach)
            self.CAS = self.CAS_from_q_c(self.q_c)
        elif CAS is not None:
            self.CAS = self.checkandsize(CAS)
            self.q_c = self.q_c_from_CAS(self.CAS)
            self.mach = self.M_inf_from_q_c(self.q_c)
            self.TAS = self.TAS_from_M_inf(self.mach)
            self.EAS = self.EAS_from_TAS(self.TAS, self.mach)
        elif EAS is not None:
            self.EAS = self.checkandsize(EAS)
            self.mach = self.M_inf_from_EAS(self.EAS)
            self.TAS = self.TAS_from_M_inf(self.mach)
            self.q_c = self.q_c_from_M_inf(self.mach)
            self.CAS = self.CAS_from_q_c(self.q_c)
        else:
            raise TypeError("Input mach, TAS, CAS, or EAS")
        self.q_inf = self.q_inf_from_TAS(self.TAS)

        # Check that computations are valid
        M = _atleast_1d(self.mach)
        self._mach_min = 0 * dimless
        self._mach_max = 1 * dimless
        if (M < self._mach_min).any() or (self._mach_max < M).any():
            raise ValueError(
                f"Mach number is out of bounds "
                f"({self._mach_min:.5g} < mach < {self._mach_max:.5g})"
                )

    def __repr__(self):
        """Output string representation of class object. Default for __str__
        :returns: string output

        """
        return self.tostring(full_output=False, imperial_units=False)

    @staticmethod
    def isentropic_mach(p_0, p):
        """Isentropic flow equation for Mach number

        :p_0: stagnation pressure
        :p: static pressure
        :returns: Mach number

        """
        y = Phys.gamma_air
        M = sqrt((2/(y-1))*((p_0/p + 1)**((y-1)/y) - 1))
        return M

    @staticmethod
    def isentropic_stagnation_pressure(M, p):
        """Isentropic flow equation for stagnation pressure

        :M: Mach number
        :p: static pressure
        :returns: stagnation pressure

        """
        y = Phys.gamma_air
        p_0 = p*(-1 + (1 + ((y-1)/2)*M**2)**(y/(y-1)))
        return p_0

    @staticmethod
    def mach_number(U, a):
        """Compute Mach number

        :U: airspeed
        :a: sound speed
        :returns: Mach number

        """
        return U/a

    @staticmethod
    def mach_airspeed(M, a):
        """Compute airspeed from Mach number

        :M: Mach number
        :a: sound speed
        :returns: airspeed

        """
        return M*a

    def checkandsize(self, inpvar):
        """Check that input is correctly typed, then size input array. If
        scalar, leave as scalar.

        :inpvar: input array (or scalar)
        :returns: sized array (or scalar)

        """
        check_dimensioned(inpvar)
        if shape(self.altitude):  # if altitude is an array
            if shape(inpvar):  # if inpvar is an array
                if inpvar.size > self.altitude.size:
                    raise TypeError("Input airspeed array size must be less "
                                    "than or equal to the altitude array "
                                    "size.")

            sizedarr = ones(shape(self.altitude))*inpvar
        else:
            sizedarr = inpvar

        return sizedarr

    def tostring(self, full_output=False, imperial_units=False):
        """String representation of data structure.

        :full_output: set to True for full output
        :imperial_units: set to True for Imperial units
        :returns: string representation

        """
        atm_str = self.atm.tostring(full_output, imperial_units)
        atm_long_str = self.atm.tostring(full_output, imperial_units)

        M_str = f"mach = {self.mach:8.5g~P}"
        if imperial_units:
            TAS_str = f"TAS = {self.TAS.to('ft/s'):8.5g~P}"
            CAS_str = f"CAS = {self.CAS.to('ft/s'):8.5g~P}"
            EAS_str = f"EAS = {self.EAS.to('ft/s'):8.5g~P}"
            q_str = f"q_inf = {self.q_inf.to('lbf/ft^2'):8.5g~P}"
        else:  # SI units
            TAS_str = f"TAS = {self.TAS.to('m/s'):8.5g~P}"
            CAS_str = f"CAS = {self.CAS.to('m/s'):8.5g~P}"
            EAS_str = f"EAS = {self.EAS.to('m/s'):8.5g~P}"
            q_str = f"q_inf = {self.q_inf.to('Pa'):8.5g~P}"

        if full_output:
            repr_str = (f"{TAS_str}\n{CAS_str}\n{EAS_str}\n{M_str}\n{q_str}\n"
                        f"{atm_long_str}")
        else:
            repr_str = f"{TAS_str}\n{CAS_str}\n{EAS_str}\n{atm_str}"
        return repr_str

    @to_base_units_wrapper
    def M_inf_from_TAS(self, TAS):
        """Compute Mach number from true airspeed.

        :TAS: true airspeed
        :returns: Mach number

        """
        a_inf = self.atm.a
        mach = __class__.mach_number(U=TAS, a=a_inf)
        return mach

    @to_base_units_wrapper
    def TAS_from_M_inf(self, mach):
        """Compute true airspeed from Mach number.

        :mach: Mach number
        :returns: true airspeed

        """
        a_inf = self.atm.a
        TAS = __class__.mach_airspeed(M=mach, a=a_inf)
        return TAS

    @to_base_units_wrapper
    def M_inf_from_EAS(self, EAS):
        """Computer Mach number from equivalent airspeed.

        :EAS: equivalent airspeed
        :returns: Mach number

        """
        a_inf_h0 = self._atm0.a
        mach = EAS/(a_inf_h0*sqrt(self.delta))
        return mach

    @to_base_units_wrapper
    def EAS_from_M_inf(self, mach):
        """Computer Mach number from equivalent airspeed.

        :mach: Mach number
        :returns: equivalent airspeed

        """
        a_inf_h0 = self._atm0.a
        EAS = mach*(a_inf_h0*sqrt(self.delta))
        return EAS

    @to_base_units_wrapper
    def q_inf_from_TAS(self, TAS):
        """Compute dynamic pressure from true airspeed.

        :TAS: true airspeed
        :returns: dynamic pressure

        """
        rho_inf = self.atm.rho
        q_inf = 0.5*rho_inf*TAS**2
        return q_inf

    @to_base_units_wrapper
    def q_c_from_CAS(self, CAS):
        """Compute impact pressure from calibrated airspeed (accounting for
           compressibility).

        :CAS: calibrated airspeed
        :returns: impact pressure

        """
        a_h0 = self._atm0.a
        p_h0 = self._atm0.p
        # Account for compressibility with the isentropic flow equation
        M_ = __class__.mach_number(U=CAS, a=a_h0)
        q_c = __class__.isentropic_stagnation_pressure(M=M_, p=p_h0)
        return q_c

    @to_base_units_wrapper
    def CAS_from_q_c(self, q_c):
        """Compute calibrated airspeed from impact pressure (accounting for
           compressibility).

        :CAS: calibrated airspeed
        :returns: impact pressure

        """
        a_h0 = self._atm0.a
        p_h0 = self._atm0.p
        # Account for compressibility with the isentropic flow equation
        M_ = __class__.isentropic_mach(p_0=q_c, p=p_h0)
        CAS = __class__.mach_airspeed(M_, a_h0)
        return CAS

    @to_base_units_wrapper
    def M_inf_from_q_c(self, q_c):
        """Compute Mach number from impact pressure.

        :q_c: impact pressure
        :returns: Mach number

        """
        p_inf = self.atm.p
        # Isentropic flow equation
        mach = __class__.isentropic_mach(p_0=q_c, p=p_inf)

        return mach

    @to_base_units_wrapper
    def q_c_from_M_inf(self, mach):
        """Compute impact pressure from Mach number.

        :mach: Mach number
        :returns: impact pressure

        """
        p_inf = self.atm.p
        # Solve for impact pressure from isentropic flow equation:
        q_c = __class__.isentropic_stagnation_pressure(M=mach, p=p_inf)
        return q_c

    @to_base_units_wrapper
    def EAS_from_TAS(self, TAS, mach):
        """Convert airspeed to true calibrated airspeed.

        :TAS: true airspeed
        :mach: Mach number
        :returns: equivalent airspeed

        """
        a_inf_h0 = self._atm0.a
        EAS = mach*(a_inf_h0*sqrt(self.delta))
        return EAS

    @to_base_units_wrapper
    def redim_force(self, S_ref):
        """Factor to redimensionalize force from force coefficient e.g.,

            .. math::

                L & = C_L q_\\infty S_ref
                  & = C_L \\text{redim_force}

        :S_ref: reference area
        :returns: redimensionalization factor

        """
        check_area_dimensioned(S_ref)
        return self.q_inf * S_ref

    @to_base_units_wrapper
    def redim_moment(self, S_ref, ell):
        """Factor to redimensionalize moment from moment coefficient e.g.,

            .. math::

                M & = C_M q_\\infty S_ref ell
                  & = C_M \\text{redim_moment}

        :S_ref: reference area
        :ell: reference length
        :returns: redimensionalization factor

        """
        check_area_dimensioned(S_ref)
        check_length_dimensioned(ell)
        return self.q_inf * S_ref * ell

    @to_base_units_wrapper
    def reynolds_number(self, ell):
        """Compute Reynolds number for the flight condition.

        :ell: length scale
        :returns: Reynolds number

        """
        nu = self.atm.nu
        TAS = self.TAS
        Re_ell = TAS*ell/nu
        return Re_ell

    def reynolds_number_per_unit_length(self, length_unit='in'):
        """Compute Reynolds number divided by length unit.

        Reynolds number is,
            Re = U_inf * ell / nu

        For some fluid dynamics solvers the user must input the Reynolds number
        in terms of Reynolds number per-unit-length,
            Re_by_ell = Re_ell / ell_in_grid_units
                      = U_inf * (ell/ell_in_grid_units) / nu

        Assume the length scale is 5 feet.  When grid units are the same as
        standard length units, such as feet vs. feet, the term
        (ell_in_grid_units/ell) is unity, and does not change the Reynolds
        number magnitude.  For example,
            ell/ell_in_grid_units = (5 ft)/(5 ft) = 1 ft/ft

        When grid units differ from standard length units, such as inches vs.
        feet, the term (ell_in_grid_units/ell) becomes,
            ell/ell_in_grid_units = (5 ft)/(60 in) = 1/12 ft/in
        which is the conversion factor between feet and inches.

        So, the Re_by_ell term becomes
            Re_by_ell = Re_ell * (1/12 1/in)

        In this case, it is helpful to compute Reynolds number per-unit-length
        in inches (length_unit='in').


        :length_unit: desired length unit as string, ('in', 'mm', 'cm')
        :returns: Reynolds number

        """
        nu_inf = self.atm.nu
        TAS = self.TAS
        Re_by_length_unit = TAS/nu_inf
        Re_by_length_unit.ito(f"1/{length_unit}")
        return Re_by_length_unit



#!/usr/bin/env python
"""
Setup file for aeroutils.
"""

from setuptools import setup, find_packages

# Safely load __version__ without importing package
with open("src/aeroutils/version.py") as f:
    exec(f.read())

if __name__ == "__main__":
    with open('README.rst', 'r') as file:
        long_description_ = file.read()

    print("Packages found:", find_packages(where='src'))  # -v option to see
    setup(
        name='aeroutils',
        version=__version__,
        packages=find_packages(where='src'),
        package_dir={'aeroutils': "src/aeroutils"},
        description="Importable utilities for aerospace problem solving.",
        long_description=long_description_,
        license="MIT License",
        url="https://github.com/MattCJones/aeroutils",
        author="Matthew C. Jones",
        author_email='matt.c.jones.aoe@gmail.com',
        scripts=None,
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Topic :: Scientific/Engineering",
            "Topic :: Utilities",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            ],
        keywords="utility aerospace engineering design problem solving",
        install_requires=('pint', 'numpy'),
    )



"""
Aerospace utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from .flightcondition import FlightCondition
from .atmosphere import Atmosphere
from .units import unit, dimless, printv

from .version import __version__
